"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const lodash_1 = require("lodash");
const create_app_1 = require("./create-app");
const types_1 = require("./types");
const Module = require('module');
const noop = () => { };
const getPermissions = (middlewares) => middlewares
    .filter(types_1.isMiddleware)
    .filter(types_1.isPermissionMiddleware)
    .filter(m => !!m.$permission)
    .map(m => m.$permission);
/**
 * Get information about where the endpoint is defined in the source code.
 */
const getEndpointSource = () => {
    const stack = new Error().stack;
    if (!stack) {
        return;
    }
    // Select the stack frame where the endpoint is defined.
    // Assume V8 stack trace format (https://v8.dev/docs/stack-trace-api):
    // [0]: Error description
    // [1]: This function call
    // [2]: MetadataApp.[method]() call
    // [3]: The target frame
    const frame = stack.split('\n')[3];
    // Match the filename and line number from a string formatted like:
    //
    //     at fn (/path/to/endpoints.js:72:11)
    //
    const result = frame.match(/^.*\((.+):(\d+):\d+\)$/);
    if (!result) {
        return;
    }
    const filename = result[1];
    const lineNumber = Number(result[2]);
    return { filename, lineNumber };
};
const isString = (a) => typeof a === 'string';
const isNumber = (a) => typeof a === 'number';
class MetadataApp {
    constructor() {
        // express methods
        this.all = noop; // All method is not supported in this stub
        this.disable = noop;
        this.disabled = noop;
        this.enable = noop;
        this.enabled = noop;
        this.engine = noop;
        this.listen = noop;
        this.METHOD = noop;
        this.options = noop;
        this.param = noop;
        this.path = noop;
        this.render = noop;
        this.route = noop;
        this.set = noop;
        // non-express methods
        this.permissions = [];
        this.response = { end: noop };
        this.responseSchema = undefined;
        this.routes = [];
        this.schema = undefined;
        this.delete = this.generateRoute('delete');
        this.get = this.generateRoute('get');
        this.patch = this.generateRoute('patch');
        this.post = this.generateRoute('post');
        this.put = this.generateRoute('put');
        this.use = (...args) => this.generateChild(...args);
    }
    generateRoute(method) {
        return (path, ...args) => {
            const [summary, description = ''] = args.filter(isString);
            const other = lodash_1.flattenDeep(args).filter(a => !!a);
            const schemas = other.filter(types_1.isSchema());
            const schema = schemas.find(s => s.$scope !== 'response');
            const responseSchema = schemas.find(s => s.$scope === 'response');
            const successStatus = other.find(isNumber) || 200;
            const middlewares = other.filter(types_1.isMiddleware);
            const deprecated = middlewares.find(m => !!m.$deprecated);
            const permissions = this.permissions.concat(getPermissions(middlewares));
            const source = getEndpointSource();
            this.routes.push({
                deprecated: deprecated && deprecated.$deprecated,
                description,
                method,
                path: path.toString(),
                permissions,
                responseSchema,
                schema: (this.schema || schema)
                    ? Object.assign({}, this.schema, schema)
                    : undefined,
                source,
                successStatus,
                summary,
            });
        };
    }
    generateChild(...args) {
        const hasPrefix = typeof args[0] === 'string';
        const prefix = hasPrefix ? args[0] : '';
        const middlewares = lodash_1.flattenDeep(!hasPrefix
            ? args
            : args.slice(1));
        const docs = middlewares
            .filter(m => m instanceof MetadataApp);
        const schemas = middlewares
            .filter(m => !(m instanceof MetadataApp) && typeof m === 'object');
        if (schemas.length) {
            this.schema = Object.assign({}, this.schema, ...schemas);
        }
        const routeLists = docs.map(d => d.routes.map(r => (Object.assign({}, r, { path: `${prefix}${r.path}`.replace(/\/$/, '') }))));
        this.routes.push(...lodash_1.flattenDeep(routeLists));
        this.permissions.push(...getPermissions(middlewares));
    }
    sort() {
        this.routes = lodash_1.sortBy(this.routes, ['path', 'method']);
    }
    getRoutes() {
        this.sort();
        return this.routes;
    }
}
// For backward compatibility, we need to monkeypatch the `require` function
// This should be removed in the next major version release of async-app
function deprecatedExpressMonkeypatch() {
    const originalRequire = Module.prototype.require;
    Module.prototype.require = function require(path) {
        // Monkeypatch require to return an `App` instead of an express instance.
        if (path === 'express') {
            const ctor = () => new MetadataApp();
            Object.assign(ctor, {
                default: ctor,
                static: noop,
            });
            return ctor;
        }
        // For all the other modules, we just don't patch anything.
        return originalRequire.apply(this, arguments); // tslint:disable-line
    };
}
const analyzeApp = (getApp) => {
    try {
        // TODO: remove this in the future
        deprecatedExpressMonkeypatch();
        // We setup the stub
        create_app_1.setProvider(() => new MetadataApp());
        // and then analyze the app by getting it's routes.
        const app = getApp();
        return app.getRoutes();
    }
    finally {
        create_app_1.resetProvider();
    }
};
exports.analyzeFile = (absolutePathToApp) => analyzeApp(() => require(absolutePathToApp));
exports.removeScope = (schema) => lodash_1.omit(schema, '$scope');
exports.default = analyzeApp;
//# sourceMappingURL=analyze.js.map