"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const custom_response_1 = require("./custom-response");
const error_1 = require("./error");
const types_1 = require("./types");
const DEFAULT_STATUS_CODE = 200;
const copyDecorators = (src, dest) => {
    const keys = Object
        .keys(src)
        .filter(k => k.startsWith('$'));
    keys.forEach(k => dest[k] = src[k]);
};
const mapMiddleware = (middleware, options, errorHandler) => {
    const fn = async (req, res, next) => {
        try {
            // First we run the middleware as usual
            const promiseOrVoid = types_1.isAsyncMiddleware(middleware)
                ? middleware(req)
                : middleware(req, res, next);
            const promise = types_1.isPromise(promiseOrVoid)
                ? promiseOrVoid
                : Promise.resolve(promiseOrVoid);
            const result = await promise;
            const isAsync = types_1.isAsyncMiddleware(middleware);
            const mappedVal = options.mapAsyncResultFn
                ? await options.mapAsyncResultFn(result, {
                    isAsyncMiddleware: isAsync,
                    isLastMiddleware: options.isLastMiddleware,
                    req,
                })
                : result;
            let val = mappedVal;
            let isRaw = false;
            let headers;
            if (mappedVal instanceof custom_response_1.CustomResponse) {
                val = mappedVal.value;
                isRaw = mappedVal.isRaw;
                headers = mappedVal.headers;
            }
            if (!isAsync) {
                // This middleware has `res` so we it will take care of the response
                return;
            }
            if (!options.isLastMiddleware || req.method === 'USE') {
                // `USE` middlewares never return stuff, so there is no need for us to
                // attempt to process the result. `options` is only passed to the last
                //  middleware, if this is not the last middleware, continue with next
                // one
                return next();
            }
            // The middleware already responded so there's nothing for us to do
            if (res.headersSent)
                return;
            if (options.validateSchema) {
                // Lets validate that schema
                const schemaErrors = options.validateSchema(val);
                if (schemaErrors.length) {
                    const response = {
                        error: 'INVALID_SCHEMA_RESPONSE',
                        path: schemaErrors[0].key,
                        source: 'response',
                    };
                    res.status(400).send(response);
                    return;
                }
            }
            if (headers) {
                res.set(headers);
            }
            res.status(options.statusCode);
            if (isRaw) {
                res.send(val);
            }
            else {
                res.json(val || {});
            }
        }
        catch (err) {
            if (!err || !(err instanceof error_1.CustomError)) {
                return next(err);
            }
            const { statusCode, error, extra } = err;
            // User defined error handler for all middlewares in app
            if (errorHandler) {
                return errorHandler(err, req, res, next);
            }
            return res.status(statusCode || 500).send(Object.assign({ error }, extra));
        }
    };
    copyDecorators(middleware, fn);
    // Set the wrapper function's name to the original function's name. When using
    // observability tools that record function names, this helps identify the
    // original function.
    Object.defineProperty(fn, 'name', { value: middleware.name });
    return fn;
};
exports.default = ({ errorHandler, compileSchema, mapAsyncResultFn, } = {}) => (args, context) => {
    const statusCode = args.find(types_1.isNumber) || DEFAULT_STATUS_CODE;
    const schema = args
        .filter(types_1.isSchema())
        .find(s => s.$scope === 'response');
    const rawSchema = schema
        ? schema.$schema
            ? schema.$schema
            : schema
        : undefined;
    const validateSchema = compileSchema && rawSchema && compileSchema(rawSchema, context);
    const middlewares = args.filter(types_1.isMiddleware);
    const lastMiddleware = middlewares[middlewares.length - 1];
    // 4 argument middlewares are error handlers, and we leave them untouched
    return args.map(m => types_1.isMiddleware(m) && m.length < 4
        ? mapMiddleware(m, 
        // Check if this is the last valid middleware (not the statusCode arg)
        {
            isLastMiddleware: m === lastMiddleware,
            mapAsyncResultFn,
            statusCode,
            validateSchema,
        }, 
        // Passes a custom error handler
        errorHandler)
        : m);
};
//# sourceMappingURL=async.js.map