"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const lodash_1 = require("lodash");
const types_1 = require("./types");
const isString = (a) => typeof a === 'string';
const createProviderMap = (middlewares) => middlewares
    .filter(types_1.isProviderMiddleware)
    .reduce((acc, m) => {
    const mNeeds = (m.$requires || []).length;
    m.$provides.forEach((model) => {
        const current = acc[model];
        // We always keep the provider with least requirements
        // Note: this can be improved by keeping the provider with a shorter
        // dependency path. Right now if a -> [x, y], b -> [z] we'll keep
        // `b`, even if `z` depends on a million things and `x` and `y` don't
        // depend on anything.
        acc[model] = (!current || (current.$requires || []).length > mNeeds)
            ? m
            : current;
    });
    return acc;
}, {});
const printBrokenMiddleware = ({ $permission, $provides, $requires, }) => JSON.stringify({ $permission, $provides, $requires });
const findRequirementsAcyclic = (pmap, m, existing, context) => {
    if (existing.includes(m)) {
        throw new Error(`Error in \`${context.method} ${context.path}\`:

    Cyclic middleware dependency found!

    The cycle triggers when trying to add:
    ${printBrokenMiddleware(m)}

    To:
    ${existing.map(printBrokenMiddleware).join('\n    ')}
    `);
    }
    if (!m.$requires || !m.$requires.length)
        return [m];
    const allProviders = lodash_1.flatten(m.$requires.map(model => pmap[model]));
    const [withRequirements, withoutRequirements] = lodash_1.partition(allProviders, p => p.$requires && p.$requires.length);
    const newExisting = [
        ...existing,
        ...withoutRequirements,
        m,
    ];
    return [
        ...withoutRequirements,
        ...lodash_1.flatten(withRequirements.map(p => findRequirementsAcyclic(pmap, p, newExisting, context))),
        m,
    ];
};
const findRequirements = (pmap, m, context) => findRequirementsAcyclic(pmap, m, [], context);
exports.default = () => (args, context) => {
    // Split arguments to organize them by type
    const strings = args.filter(isString);
    const objects = args.filter(types_1.isSchema());
    const functions = args.filter(types_1.isMiddleware);
    // Split functions to handle loaders and permissions separately
    const [noOrder, middlewares] = lodash_1.partition(functions, types_1.isNonOrderable);
    // Validate that every requirement is met
    const everythingRequired = lodash_1.flatten(middlewares.map(m => m.$requires || []));
    const everythingProvided = lodash_1.flatten(middlewares.map(m => m.$provides || []));
    const missing = lodash_1.difference(everythingRequired, everythingProvided);
    if (missing.length) {
        throw new Error(`Error in \`${context.method} ${context.path}\`:
      Missing required loader for: ${missing.join(', ')}.
    `);
    }
    // In here we'll drop the ordered middlewares. We start with the ones that
    // should not be ordered.
    const orderedMiddlewares = [...noOrder];
    const addMiddleware = (m) => {
        if (!orderedMiddlewares.includes(m)) {
            orderedMiddlewares.push(m);
        }
    };
    // The goal is to add permissions as soon as possible, but in order to add
    // them, we first need to satisfy their requirements.
    const providerMap = createProviderMap(middlewares);
    // We transform each permission into a list of middlewares required to inject
    // that permission, then sort them from least to most requirements
    const permissions = middlewares
        .filter(types_1.isPermissionMiddleware)
        .map(p => findRequirements(providerMap, p, context))
        .sort((a, b) => a.length - b.length);
    // Lets do this...
    permissions.forEach((middlwareList) => {
        middlwareList.forEach(addMiddleware);
    });
    // Now that that is taken care of, we need to add all the remaining
    // middlewares and since `addMiddleware` avoids dups, we can just go ahead
    // re-add everything
    middlewares.forEach(addMiddleware);
    const ms = [
        ...strings,
        ...objects,
        ...orderedMiddlewares,
    ].filter(m => !!m);
    return ms;
};
//# sourceMappingURL=order.js.map