"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const lodash_1 = require("lodash");
const get_argument_obj_keys_1 = __importDefault(require("./get-argument-obj-keys"));
const types_1 = require("./types");
const util_1 = require("./util");
const areEqual = (arr1, arr2) => arr1.length === arr2.length && arr1.every(item => arr2.includes(item));
const checkExpectedModels = (entity, entityName, models) => {
    const expectedAll = [];
    util_1.getKeys(entity).forEach((action) => {
        const spec = entity[action];
        if (types_1.isPermissionFn(spec)) {
            expectedAll.push(...get_argument_obj_keys_1.default(spec));
        }
        if (types_1.isPermissionEntity(spec)) {
            Object.keys(spec).forEach((subaction) => {
                const subspec = spec[subaction];
                if (types_1.isPermissionFn(subspec)) {
                    expectedAll.push(...get_argument_obj_keys_1.default(subspec));
                }
            });
        }
    });
    const given = util_1.getKeys(models).filter(model => !!models[model]);
    const expected = lodash_1.uniq(expectedAll);
    if (!areEqual(given, expected)) {
        throw new Error(`Wrong expected models for "${entityName}".
      Expected: ${expected.join(', ')};
      Given: ${given.join(', ')}.`);
    }
};
const tryPermission = (permissionFn, requiredModels) => {
    try {
        return { access: permissionFn(requiredModels) };
    }
    catch (error) {
        return { access: false, error };
    }
};
const assertEntity = (entity, entityName) => {
    if (!entity)
        throw new Error(`Invalid entity ${entityName}`);
    return entity;
};
// Accepting an optional tryFunction lets users override the default try/catch
// which means you can in theory pull off the `extra` key in thrown errors etc
exports.computePermissions = (entities, entityName, requiredModels) => {
    const entity = assertEntity(entities[entityName], entityName);
    checkExpectedModels(entity, entityName, requiredModels);
    const permissions = {
        $errors: {},
    };
    Object.keys(entity).forEach((action) => {
        const spec = entity[action];
        // load permission on entire entity, e.g.: $permissions.delete = true
        if (types_1.isPermissionFn(spec)) {
            const { access, error } = tryPermission(spec, requiredModels);
            permissions[action] = access;
            if (error) {
                (permissions.$errors)[action] = error;
            }
        }
        // load subpermissions, e.g.: $permissions['delete.editHash'] = true
        if (types_1.isPermissionEntity(spec)) {
            Object.keys(spec).forEach((subaction) => {
                const permissionFn = spec[subaction];
                if (types_1.isPermissionFn(permissionFn)) {
                    const { access, error } = tryPermission(permissionFn, requiredModels);
                    const permKey = `${action}.${subaction}`;
                    permissions[permKey] = access;
                    if (error) {
                        (permissions.$errors)[permKey] = error;
                    }
                }
            });
        }
    });
    return permissions;
};
//# sourceMappingURL=compute.js.map