"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const lodash_1 = require("lodash");
const schema_1 = require("./schema");
const expandTemplates = (getAlias, s) => (s
    ? s.replace(/:(\w+)/g, (_, p) => `{${(getAlias && getAlias(p)) || p}}`)
    : s);
const defaultGeneratePermissions = (permissions) => {
    if (!permissions.length)
        return '';
    return `
  ### Permissions

  ${permissions.map(p => `- ${p}\n`).join('')}

  `;
};
const isQuerySchema = (schema, method) => schema &&
    (schema.$scope || schema_1.METHOD_SOURCE_MAP[method]) === 'query';
const wrapBodySchema = (schema) => schema
    ? [{
            in: 'body',
            name: 'body',
            required: true,
            schema,
        }]
    : undefined;
const wrapQuerySchema = (schema) => (lodash_1.isPlainObject(schema) && schema.properties)
    ? Object.entries(schema.properties)
        .map(([key, value]) => (Object.assign({ in: 'query', name: key, type: value.enum ? 'string' : value.type }, (value.enum ? { enum: value.enum } : undefined))))
    : undefined;
const generateEndpoint = ({ generatePermissions = defaultGeneratePermissions, generateSchema, getParamAlias, }, { method, path, category, deprecated, description, permissions, summary, schema, responseSchema, successStatus, }) => ({
    consumes: ['application/json'],
    deprecated,
    description: expandTemplates(getParamAlias, `${description}${generatePermissions(permissions)}`),
    operationId: `${method}${path}`.replace(/\W+/g, '-'),
    parameters: schema
        && generateSchema
        && (isQuerySchema(schema, method) ? wrapQuerySchema : wrapBodySchema)(generateSchema(schema)),
    produces: ['application/json'],
    responses: {
        [successStatus]: {
            description: 'Success',
            schema: responseSchema && generateSchema && generateSchema(responseSchema),
        },
    },
    summary: expandTemplates(getParamAlias, summary),
    tags: [category],
});
const showRoute = (allowDeprecated = 'in-use') => (r) => !r.deprecated
    || (r.deprecated === 'in-use' && allowDeprecated !== 'none')
    || (allowDeprecated === 'all');
const defaultSortCategories = (a, b) => {
    if (a === 'deprecated')
        return 1;
    if (b === 'deprecated')
        return -1;
    if (a === 'misc')
        return 1;
    if (b === 'misc')
        return -1;
    return a < b ? -1 : 1;
};
exports.default = (routes, opts = {}) => {
    const routesWithCat = routes.map((r) => {
        return Object.assign({}, r, { category: opts.getCategory && opts.getCategory(r) || 'misc' });
    });
    const groups = lodash_1.groupBy(routesWithCat, 'category');
    const [proper, misc] = lodash_1.partition(Object.values(groups), v => v.length > 1);
    const finalRoutes = [
        ...lodash_1.flattenDeep(proper),
        ...lodash_1.flattenDeep(misc).map(r => (Object.assign({}, r, { category: 'misc' }))),
    ].filter(showRoute(opts.deprecated));
    const paths = finalRoutes.reduce((acc, route) => {
        const spath = expandTemplates(opts.getParamAlias, route.path.toString());
        const pathEntry = acc[spath] || (acc[spath] = {});
        pathEntry[route.method] = generateEndpoint(opts, route);
        return acc;
    }, {});
    const cats = lodash_1.uniq(finalRoutes.map(r => r.category))
        .sort(opts.sortCategories || defaultSortCategories);
    return {
        paths,
        tags: cats.map(c => ({ name: c })),
    };
};
//# sourceMappingURL=swagger.js.map