import { Express, NextFunction, Request, Response } from 'express';
import { CustomError } from './error';
export interface Context {
    method: Method;
    path?: string;
}
export interface ErrorContext extends Context {
    req: Request;
}
export interface Entities {
    _: unknown;
}
export declare type Req<TEntities extends Entities, T extends keyof TEntities> = Pick<TEntities, T> & Request;
export declare type Decorator = {
    $provides?: string[];
    $requires?: string[];
    $permission?: string;
    $noOrder?: boolean;
    $deprecated?: 'in-use' | 'redirect' | 'rewrite';
};
export declare type CommonMiddleware<TEntities extends Entities> = ((req: Req<TEntities, keyof TEntities>, res: Response, next: NextFunction) => Promise<any> | void) & Decorator;
export declare type AsyncMiddleware<TEntities extends Entities> = ((req: Req<TEntities, keyof TEntities>) => Promise<unknown> | unknown) & Decorator;
export declare type Middleware<TEntities extends Entities> = CommonMiddleware<TEntities> | AsyncMiddleware<TEntities>;
export declare type ProviderMiddleware<TEntities extends Entities> = Middleware<TEntities> & {
    $provides: string[];
};
export declare type NonOrderableMiddleware<TEntities extends Entities> = Middleware<TEntities> & {
    $noOrder: boolean;
};
export declare type RequireMiddleware<TEntities extends Entities> = Middleware<TEntities> & {
    $requires: string[];
};
export declare type PermissionMiddleware<TEntities extends Entities> = Middleware<TEntities> & {
    $permission: string;
};
export declare const isProviderMiddleware: <TEntities extends Entities>(o: Middleware<TEntities>) => o is ProviderMiddleware<TEntities>;
export declare const isRequirMiddleware: <TEntities extends Entities>(o: Middleware<TEntities>) => o is RequireMiddleware<TEntities>;
export declare const isPermissionMiddleware: <TEntities extends Entities>(o: Middleware<TEntities>) => o is PermissionMiddleware<TEntities>;
export declare const isNonOrderable: <TEntities extends Entities>(o: Middleware<TEntities>) => o is NonOrderableMiddleware<TEntities>;
export declare type RequestScope = 'body' | 'query';
export declare type Scope = RequestScope | 'response';
export declare type Schema<TSchema = {
    [k: string]: any;
}> = {
    $schema?: TSchema;
    $scope?: Scope;
} & TSchema;
export declare type ValidationError = {
    expected?: string;
    key: (string | number)[];
};
export declare type ValidateSchema = (obj: any) => ValidationError[];
export declare type CompileSchema<T> = (schema: Schema<T>, context: Context) => ValidateSchema;
export declare type GenerateSchemaErrorFn = (errors: ValidationError[], source: string, context: ErrorContext) => any;
export declare type ErrorHandlerFn<TEntities extends Entities> = (errors: CustomError, req: Req<TEntities, keyof TEntities>, res: Response, next: NextFunction) => any;
export declare type MiddlewareArg<TEntities extends Entities> = CommonMiddleware<TEntities> | CommonMiddleware<TEntities>[] | AsyncMiddleware<TEntities>;
export declare type ArgumentOption<TEntities extends Entities, TSchema> = MiddlewareArg<TEntities> | string | number | RegExp | Schema<TSchema>;
export declare const isNumber: (m: unknown) => m is number;
export declare const isPromise: (f: unknown) => f is Promise<any>;
export declare const isSchema: <TSchema>() => (m: unknown) => m is Schema<TSchema>;
export declare const isMiddlewareArg: <TEntities extends Entities>(o: ArgumentOption<TEntities, unknown>) => o is MiddlewareArg<TEntities>;
export declare const isAsyncMiddleware: <TEntities extends Entities>(o: ArgumentOption<TEntities, unknown>) => o is AsyncMiddleware<TEntities>;
export declare const isMiddleware: <TEntities extends Entities>(o: ArgumentOption<TEntities, unknown>) => o is Middleware<TEntities>;
export declare type Converter<TEntities extends Entities, TSchema> = (m: ArgumentOption<TEntities, TSchema>[], context: Context) => ArgumentOption<TEntities, TSchema>[];
export declare type MapAsyncResultFn<TEntities extends Entities> = (value: unknown, opts: {
    isAsyncMiddleware: boolean;
    isLastMiddleware: boolean;
    req: Req<TEntities, keyof TEntities>;
}) => unknown;
export interface Opts<TEntities extends Entities, TSchema> {
    converters?: Converter<TEntities, TSchema>[];
    compileSchemaFn?: CompileSchema<TSchema>;
    generateSchemaErrorFn?: GenerateSchemaErrorFn;
    errorHandlerFn?: ErrorHandlerFn<TEntities>;
    validateResponseSchema?: boolean;
    mapAsyncResultFn?: MapAsyncResultFn<TEntities>;
}
export declare type Method = keyof AsyncApp<Entities, unknown>;
export declare type RequestHandler<TEntities extends Entities, TSchema> = (...m: ArgumentOption<TEntities, TSchema>[]) => void;
interface AsyncApp<TEntities extends Entities, TSchema> {
    delete: RequestHandler<TEntities, TSchema>;
    get: RequestHandler<TEntities, TSchema>;
    patch: RequestHandler<TEntities, TSchema>;
    post: RequestHandler<TEntities, TSchema>;
    put: RequestHandler<TEntities, TSchema>;
    use: RequestHandler<TEntities, TSchema>;
}
export declare type App<TEntities extends Entities, TSchema> = Express & AsyncApp<TEntities, TSchema>;
export {};
